#include "alloc.h"
/* mostly from dietlibc */

#if defined(__i386__) || defined(__mips__) || defined(__arm__) || defined(__powerpc__) || defined(__s390__)
# define PAGE_SIZE 4096UL
#elif defined(__alpha__) || defined(__sparc__)
# define PAGE_SIZE 8192UL
#endif

#ifndef MREMAP_MAYMOVE
# define MREMAP_MAYMOVE 1UL
#endif

typedef union {
 void *next;
 size_t size;
} mem_alloc_t;

static mem_alloc_t *__small_mem[8];

static size_t get_index(size_t _size);
static void alloc_small_free(void*_ptr,size_t _size);
static void *alloc_small_malloc(size_t _size);

#define BLOCK_START(b)  (((void*)(b))-sizeof(mem_alloc_t))
#define BLOCK_RET(b)    (((void*)(b))+sizeof(mem_alloc_t))
#define PAGE_ALIGN(s)   (((s)+PAGE_SIZE-1)&(unsigned long)(~(PAGE_SIZE-1)))

#define __SMALL_NR(i)    (PAGE_SIZE/(i))
#define __MIN_SMALL_SIZE __SMALL_NR(256) /*   16 /   32 */
#define __MAX_SMALL_SIZE __SMALL_NR(2)   /* 2048 / 4096 */
#define GET_SIZE(s)      (__MIN_SMALL_SIZE<<get_index((s)))
#define FIRST_SMALL(p)   (((unsigned long)(p))&(~(PAGE_SIZE-1)))

static void *mmap_alloc(size_t size) {
 return mmap(0, size, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, (size_t)0);
}

static inline int __ind_shift() { return (PAGE_SIZE==4096)?4:5; }

static size_t get_index(size_t _size) {
 register size_t idx=0;
 if (_size) {
  register size_t size=((_size-1)&(PAGE_SIZE-1))>>__ind_shift();
  while(size) { size>>=1; ++idx; }
 }
 return idx;
}

static void alloc_small_free(void*_ptr,size_t _size) {
 mem_alloc_t* ptr=BLOCK_START(_ptr);
 size_t size=_size;
 size_t idx=get_index(size);

 mem_set(ptr,0,size);   /* allways zero out small mem */

 ptr->next=__small_mem[idx];
 __small_mem[idx]=ptr;
}

static void *alloc_small_malloc(size_t _size) {
 mem_alloc_t *ptr;
 size_t size=_size;
 size_t idx;

 idx=get_index(size);
 ptr=__small_mem[idx];

 if (ptr==0)  {
  register int i,nr;
  ptr=mmap_alloc(PAGE_SIZE);
  if (ptr==MAP_FAILED) return MAP_FAILED;

  __small_mem[idx]=ptr;

  nr=__SMALL_NR(size)-1;
  for (i=0;i<nr;i++) {
   ptr->next=(((void*)ptr)+size);
   ptr=ptr->next;
  }
  ptr->next=0;

  ptr=__small_mem[idx];
 }

 /* get a free block */
 __small_mem[idx]=ptr->next;
 ptr->next=0;

 return ptr;
}

/* -- PUBLIC FUNCTIONS ---------------------------------------------------- */

void alloc_free(void *ptr) {
 register size_t size;
 if (ptr) {
  size=((mem_alloc_t*)BLOCK_START(ptr))->size;
  if (size) {
   if (size<=__MAX_SMALL_SIZE) {
    alloc_small_free(ptr,size);
   } else {
    munmap(BLOCK_START(ptr),size);
   }
  }
 }
}

void *alloc_malloc(size_t size) {
 mem_alloc_t *ptr;
 size_t need;
 if (size) {
  size+=sizeof(mem_alloc_t);
  if (size<=__MAX_SMALL_SIZE) {
   need=GET_SIZE(size);
   if ((ptr=alloc_small_malloc(need))==MAP_FAILED) goto err_out;
  } else {
   need=PAGE_ALIGN(size);
   if ((ptr=mmap_alloc(need))==MAP_FAILED) goto err_out;
  }
  ptr->size=need;
  return BLOCK_RET(ptr);
 }
err_out:
 draw_exit("Sorry, you're out of memory :(\n", ENOMEM);
 return 0;
}

void *alloc_realloc(void* ptr, size_t _size) {
  register size_t size=_size;
  if (ptr) {
    if (size) {
      mem_alloc_t* tmp=BLOCK_START(ptr);
      size+=sizeof(mem_alloc_t);
      size=(size<=__MAX_SMALL_SIZE)?GET_SIZE(size):PAGE_ALIGN(size);
      if (tmp->size!=size) {
	if ((tmp->size<=__MAX_SMALL_SIZE)) {
	  void *new=alloc_malloc(_size);
	  if (new) {
	    register mem_alloc_t* foo=BLOCK_START(new);
	    size=foo->size;
	    if (size>tmp->size) size=tmp->size;
	    mem_cpy(new,ptr,size-sizeof(mem_alloc_t));
	    alloc_free(ptr);
	  }
	  ptr=new;
	}
	else {
	  register mem_alloc_t* foo;
	  size=PAGE_ALIGN(size);
#ifdef MREMAP_MAYMOVE
	  foo=mremap(tmp,tmp->size,size,MREMAP_MAYMOVE);
#else
	  foo=mremap(tmp,tmp->size,size,0);
#endif
	  if (foo==MAP_FAILED) {
	    draw_exit("Sorry, you're out of memory :(\n", ENOMEM);
	    return 0;
	  }
	  else {
	    foo->size=size;
	    ptr=BLOCK_RET(foo);
	  }
	}
      }
    }
    else { /* size==0 */
      alloc_free(ptr);
    }
  }
  else { /* ptr==0 */
    if (size) {
      ptr=alloc_malloc(size);
    }
  }
  return ptr;
}
